from logging import getLogger
from opentelemetry.sdk.resources import (
    Resource,
    ResourceDetector,
    ProcessResourceDetector,
    OTELResourceDetector,
    OsResourceDetector,
    get_aggregated_resources,
)
from opentelemetry_resourcedetector_docker import DockerResourceDetector
from middleware.options import MWOptions
from middleware.detectors.detector import process_detector_input, get_detectors
from middleware.version import __version__
import typing

_logger = getLogger(__name__)


def create_resource(options: MWOptions):
    """
    Configures and returns a new OpenTelemetry Resource.

    Args:
        options (MWOptions): the middleware options to configure with
        resource (Resource): the resource to use with the new resource

    Returns:
        Resource: the new Resource
    """
    attributes = {
        "service.name": options.service_name,
        "mw.sdk.version": __version__,
        "mw.app.lang": "python",
        "runtime.metrics.python": "true",
    }
    if "https" in options.target:
        attributes["mw.serverless"] = "true"
    if options.access_token is not None:
        attributes["mw.account_key"] = options.access_token
    if options.project_name is not None:
        attributes["project.name"] = options.project_name
    if options.custom_resource_attributes is not None:
        try:
            extra = {
                k.strip(): v.strip()
                for k, v in (
                    item.split("=")
                    for item in options.custom_resource_attributes.split(",")
                )
            }
            attributes.update(extra)
        except Exception as exc:
            _logger.warning(
                "Skipped Custom attributes: parsing error expected format `abcd=1234,wxyz=5678`"
            )

    detectors: typing.List["ResourceDetector"] = [
        OTELResourceDetector(),
        ProcessResourceDetector(),
        OsResourceDetector(),
        DockerResourceDetector(),
    ]

    if options.detectors is not None:
        extra_detectors = get_detectors(process_detector_input(options.detectors))
        if extra_detectors is not None and len(extra_detectors) > 0:
            detectors.extend(extra_detectors)
    resources = get_aggregated_resources(
        detectors,
        initial_resource=Resource.create(attributes),
    )
    return resources



from middleware.distro import mw_tracker, record_exception
from typing import Collection
import sys
import threading
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from opentelemetry import trace
from typing import Optional
from middleware.options import (
    MWOptions,
    DETECT_AWS_BEANSTALK,
    DETECT_AWS_EC2,
    DETECT_AWS_ECS,
    DETECT_AWS_EKS,
    DETECT_AWS_LAMBDA,
    DETECT_AZURE_APP_SERVICE,
    DETECT_AZURE_FUNCTIONS,
    DETECT_AZURE_VM,
    DETECT_GCP,
    DETECT_ENVVARS,
)

__all__ = [
    "mw_tracker",
    "record_exception",
    "MWOptions",
    "DETECT_AWS_BEANSTALK",
    "DETECT_AWS_EC2",
    "DETECT_AWS_ECS",
    "DETECT_AWS_EKS",
    "DETECT_AWS_LAMBDA",
    "DETECT_AZURE_APP_SERVICE",
    "DETECT_AZURE_FUNCTIONS",
    "DETECT_AZURE_VM",
    "DETECT_GCP",
    "DETECT_ENVVARS",
]

tracer = trace.get_tracer(__name__)

# Hook for threading exceptions (Python 3.8+)
def thread_excepthook(args):
    record_exception(args.exc_type, args.exc_value, args.exc_traceback)

class ExceptionInstrumentor(BaseInstrumentor):
    def instrumentation_dependencies(self) -> Collection[str]:
        """Return dependencies if this instrumentor requires any."""
        return []
    
    def _instrument(self, **kwargs):
        """Automatically sets sys.excepthook when the instrumentor is loaded."""
        sys.excepthook = record_exception
        threading.excepthook = thread_excepthook

    def _uninstrument(self, **kwargs):
        """Restores default sys.excepthook if needed."""
        sys.excepthook = sys.__excepthook__

# Load the instrumentor
ExceptionInstrumentor().instrument()

# Automatic exception handling for flask
try:
    from flask import Flask, request
    from flask.signals import got_request_exception, appcontext_pushed
except ImportError:
    Flask = None
    got_request_exception = None
    appcontext_pushed = None

if Flask and got_request_exception and appcontext_pushed:
    def _capture_exception(sender, exception, **extra):
        exc_type, exc_value, exc_traceback = sys.exc_info()  # Get exception details
        if exc_type and exc_value and exc_traceback:
            record_exception(exc_type, exc_value, exc_traceback)
        else:
            print("Unable to capture exception details.")

    def try_register_flask_handler(app: Flask):
        """Registers the exception handler using Flask signals."""
        got_request_exception.connect(_capture_exception, app)
        print("✅ Flask error handler registered via signal.")

    def _auto_register(sender, **extra):
        """Automatically registers the handler when a Flask app context is pushed."""
        try_register_flask_handler(sender)

    # Connect to Flask's appcontext_pushed to register automatically
    appcontext_pushed.connect(_auto_register)


# Automatic exception handling for FastAPI
try:
    from starlette.middleware.base import BaseHTTPMiddleware
    from starlette.requests import Request
    from starlette.responses import JSONResponse
    from fastapi import FastAPI
except ImportError:
    BaseHTTPMiddleware = None
    Request = None
    JSONResponse = None
    FastAPI = None

if BaseHTTPMiddleware and FastAPI:
    class ExceptionMiddleware(BaseHTTPMiddleware):
        """Middleware to catch unhandled exceptions globally."""
        async def dispatch(self, request: Request, call_next):
            try:
                return await call_next(request)
            except Exception as exc:
                exc_type, exc_value, exc_traceback = exc.__class__, exc, exc.__traceback__
                record_exception(exc_type, exc_value, exc_traceback)
                
                return JSONResponse(
                    status_code=500,
                    content={"detail": "Internal Server Error"},
                )
            
    from fastapi import FastAPI
    # from starlette.middleware import ExceptionMiddleware

    _original_init = FastAPI.__init__

    def new_fastapi_init(self, *args, **kwargs):
        _original_init(self, *args, **kwargs)
        print("✅ FastAPI instance created, registering ExceptionMiddleware.")
        self.add_middleware(ExceptionMiddleware)

    FastAPI.__init__ = new_fastapi_init


from logging import getLogger
from typing import Optional

from opentelemetry.sdk.trace.sampling import (
    ALWAYS_OFF,
    ALWAYS_ON,
    TraceIdRatioBased,
    Sampler,
    SamplingResult,
)

from opentelemetry.trace import Link, SpanKind
from opentelemetry.trace.span import TraceState
from opentelemetry.util.types import Attributes
from opentelemetry.context import Context

from middleware.options import DEFAULT_SAMPLE_RATE, MWOptions

_logger = getLogger(__name__)


def configure_sampler(
    options: Optional[MWOptions] = None,
):
    """Configures and returns an OpenTelemetry Sampler that is
    configured based on the sample_rate determined in MWOptions.
    The configuration initializes a DeterministicSampler with
    an inner sampler of either AlwaysOn (1), AlwaysOff (0),
    or a TraceIdRatio as 1/N.

    These samplers do not take into account the parent span's
    sampling decision.

    Args:
        options (MWOptions): the MWOptions containing
        sample_rate used to configure the deterministic sampler.

    Returns:
        DeterministicSampler: the configured Sampler based on sample_rate
    """
    if options is None:
        return DeterministicSampler(DEFAULT_SAMPLE_RATE)

    return DeterministicSampler(options.sample_rate)


class DeterministicSampler(Sampler):
    """Implementation of :class:`Sampler` that uses an inner sampler
    of either AlwaysOn (1), AlwaysOff (0), or a TraceIdRatio as 1/N
    to determine a SamplingResult and SamplingDecision for a given span
    in a trace. We append a SampleRate attribute to the span with the
    given sample rate.

    Note: These samplers do not take into account the parent span's
    sampling decision.
    """

    def __init__(self, rate: int):
        self.rate = rate

        if self.rate <= 0:
            # Sampler that never samples spans, regardless of the
            # parent span's sampling decision
            self._sampler = ALWAYS_OFF

        elif self.rate == 1:
            # Sampler that always samples spans, regardless of the
            # parent span's sampling decision
            self._sampler = ALWAYS_ON

        else:
            # Sampler that samples probabilistically based on rate..
            ratio = 1.0 / self.rate
            self._sampler = TraceIdRatioBased(ratio)

    # pylint: disable=too-many-arguments
    def should_sample(
        self,
        parent_context: Context,
        trace_id: int,
        name: str,
        kind: SpanKind = None,
        attributes: Attributes = None,
        links: Link = None,
        trace_state: "TraceState" = None,
    ) -> "SamplingResult":
        # append our SampleRate field to attributes
        sample_rate = {"SampleRate": self.rate}
        if attributes is None:
            attributes = sample_rate
        else:
            attributes.update(sample_rate)
        # using _sampler logic based on rate (OFF, ON, TraceIdRatio)
        return self._sampler.should_sample(
            parent_context, trace_id, name, kind, attributes, links, trace_state
        )

    def get_description(self) -> str:
        return "MWDeterministicSampler"



import grpc
import sys
import logging
from opentelemetry.sdk.resources import Resource
from opentelemetry.exporter.otlp.proto.grpc._log_exporter import (
    OTLPLogExporter,
)
from opentelemetry.sdk._logs import LoggerProvider, LoggingHandler
from opentelemetry.sdk._logs.export import (
    BatchLogRecordProcessor,
    SimpleLogRecordProcessor,
    ConsoleLogExporter,
)
from opentelemetry._logs import set_logger_provider
from logging import LogRecord
from middleware.options import MWOptions, log_levels

_logger = logging.getLogger(__name__)


def create_logger_handler(options: MWOptions, resource: Resource) -> LoggingHandler:
    """
    Configures and returns a new LoggerProvider to send logs telemetry.

    Args:
        options (MWOptions): the middleware options to configure with
        resource (Resource): the resource to use with the new logger provider

    Returns:
        LoggerProvider: the new logger provider
    """
    exporter = OTLPLogExporter(
        endpoint=options.target,
        compression=grpc.Compression.Gzip,
    )
    logger_provider = LoggerProvider(resource=resource, shutdown_on_exit=True)
    logger_provider.add_log_record_processor(BatchLogRecordProcessor(exporter))
    if options.console_exporter:
        if options.console_exporter:
            output = sys.stdout
            if options.debug_log_file:
                log_file = "mw-logs"
                try:
                    output = open(log_file, "w")
                except Exception:
                    _logger.error(f"Cannot open the log file for writing: {log_file}")
                    output = sys.stdout
        logger_provider.add_log_record_processor(
            SimpleLogRecordProcessor(
                ConsoleLogExporter(
                    out=output,
                )
            )
        )

    handler = MWLoggingHandler(
        level=log_levels[options.log_level], logger_provider=logger_provider
    )
    set_logger_provider(logger_provider)

    return handler

class MWLoggingHandler(LoggingHandler):
    @staticmethod
    def _get_attributes(record: LogRecord):
        attributes = LoggingHandler._get_attributes(record)

        for key, value in attributes.items():
            if key == "request":
                if hasattr(value, "method") and hasattr(value, "path"):
                    if len(vars(value)) == 2:
                        attributes[key] = f'{value.method} {value.path}'
                    else:
                        attributes[key] = str(value)
                else:
                    attributes[key] = str(value)
            elif not isinstance(value, (bool, str, bytes, int, float)):
                attributes[key] = str(value)

        return attributes